global mread, app_file, rbones, ani_file, anims, app_geos, cmeshes, fhardpoints
global num_bones, g_num_geos, g_num_subobjects, tag_size, tag_offset, cur_pos, cur_frame, f_num_anims, g_num_anims, num_meshes
global num_hardpoints

struct app_prstransform (
	rot, pos, scale
)

struct ani_frame_pos (
	key, pos
)

struct ani_frame_rot (
	key, rot
)

struct ani_frame_scale (
	key, scale
)

struct ani_bone (
	name, num_keys_pos, num_keys_rot, num_keys_scale, frames_pos, frames_rot, frames_scale
)

struct animation (
	name, num_bones, ani_bones, num_frames, applyed
)

struct app_bone (
	name, parentID, trans, maxbone, bmat
)

struct app_geoset (
	num_subobjects, subobjects
)

struct app_subobject (
	name, num_verts, verts, infl, num_faces, indices, bbox
)

struct app_vert (
	x, y, z, nx, ny, nz, nw, unk1, unk2, u, v, unk3,  tx, ty, tz, tw, bix, biy, biz, biw, unk4
)

struct app_influence (
	boneids, weights
)

struct app_hardpoint (
	name, parentID, pos, rot
)

fn echo msg = (
	format "%\n" (msg) to:listener
)

fn app_reset_globals = (
	mread = app_file = rbones = afile = undefined
	num_bones = tag_size = tag_offset = cur_pos = cur_frame = f_num_anims = g_num_anims = num_meshes = 0
	cur_frame = 1
	gc()
)

fn app_read_fixed_str fixedLen stream = (
	local str = ""
	for i = 1 to fixedLen do
	(
		str += bit.intAsChar (ReadByte stream #unsigned)
	)
	return str
)

fn app_open_file filename = (
	stream = fopen filename "rb"
	if stream == undefined then(
		print "app_file not found "+filename
		throw "app_file not found"
	)
	return stream
)

fn app_close_file stream = (
	fclose stream
)

fn app_skip_byte count stream = (
	local unknown
	case count of
	(
		2: unknown = readshort stream #unsigned
		4: unknown = readlong stream #unsigned
		default:
		(
			for i = 1 to count do
			(
				unknown = readbyte stream #unsigned
			)
		)
	)
)

fn app_file_pos offset stream = (
	local cur_pos = ftell stream
	fseek stream (offset + 16) #seek_set
	return cur_pos
)

fn app_file_restore_pos pos stream = (
	fseek stream pos #seek_set
)

fn app_read_bones = (
	rbones = #()
	rbones[num_bones] = 0
	for i = 1 to num_bones do
	(
		local cb = app_bone()
		cb.name = app_read_fixed_str 64 app_file
		cb.parentID = readlong app_file #signed + 1
		app_skip_byte 24	app_file	--two vector3 probably bone bounding box
		app_skip_byte 16 app_file	--one vector3 + float probably bounding sphere
		cb.trans = #()
		cb.trans[4] = 0
		for j = 1 to 3 do(
			local pt = app_prstransform()
			pt.rot = quat 0.0 0.0 0.0 1.0
			pt.rot.x = readfloat app_file
			pt.rot.y = readfloat app_file
			pt.rot.z = readfloat app_file
			pt.rot.w = readfloat app_file
			pt.pos = point3 0.0 0.0 0.0
			pt.pos.x = readfloat app_file
			pt.pos.y = readfloat app_file
			pt.pos.z = readfloat app_file
			pt.scale = readfloat app_file
			cb.trans[j] = pt
		)
		app_skip_byte 4 app_file	--num_collision_shapes
		app_skip_byte 4 app_file	--collision_shapes_offset
		app_skip_byte 4 app_file	--collision_shapes_size
		app_skip_byte 4 app_file	--pad0
		app_skip_byte 4 app_file	--constraints_offset
		app_skip_byte 4 app_file	--constraints_size
		app_skip_byte 4 app_file	--pad1
		app_skip_byte 4 app_file	--particle_system_sno
		rbones[i] = cb
	)
)

fn app_debug_print_bones = (
	for i = 1 to num_bones do(
		local cb = rbones[i]
		print "name"
		print cb.name
		print "parentID"
		print cb.parentID
		print "pos"
		print cb.trans[1].pos.x
		print cb.trans[1].pos.y
		print cb.trans[1].pos.z
		print "rot"
		eu = quattoeuler cb.trans[1].rot
		print eu.x
		print eu.y
		print eu.z
		print "scale"
		print cb.trans[1].scale
	)
)

fn app_read_subobjects count = (
	subs = #()
	subs[count] = 0
	for i = 1 to count do(
		app_skip_byte 4 app_file 	--unknown
		s = app_subobject()
		s.num_verts = readlong app_file #unsigned
		s.verts = #()
		s.verts[s.num_verts] = 0
		offset = readlong app_file #unsigned
		size = readlong app_file #unsigned
		pos = app_file_pos offset app_file
		for j = 1 to s.num_verts do (
			v = app_vert()
			v.x = readfloat app_file
			v.y = readfloat app_file
			v.z = readfloat app_file
			v.nx = readbyte app_file #unsigned
			v.ny = readbyte app_file #unsigned
			v.nz = readbyte app_file #unsigned
			v.nw = readbyte app_file #unsigned
			v.unk1 = readlong app_file #unsigned
			v.unk2 = readlong app_file #unsigned
			v.u = readshort app_file #unsigned
			v.v = readshort app_file #unsigned
			v.unk3 = readlong app_file #unsigned
			v.tx = readbyte app_file
			v.ty = readbyte app_file
			v.tz = readbyte app_file
			v.tw = readbyte app_file
			v.bix = readbyte app_file
			v.biy = readbyte app_file
			v.biz = readbyte app_file
			v.biw = readbyte app_file
			v.unk4 = readlong app_file #unsigned
			s.verts[j] = v
		)
		app_file_restore_pos pos app_file
		app_skip_byte 4 app_file	--pad
		s.infl = #()
		s.infl[s.num_verts] = 0
		offset = readlong app_file #unsigned
		size = readlong app_file #unsigned
		pos = app_file_pos offset app_file
		for j = 1 to s.num_verts do(
			inf = app_influence()
			inf.boneids = #()
			inf.boneids[3] = 0
			inf.weights = #()
			inf.weights[3] = 0
			inf.boneids[1] = readlong app_file #unsigned +1
			inf.weights[1] = readfloat app_file
			inf.boneids[2] = readlong app_file #unsigned + 1
			inf.weights[2] = readfloat app_file
			inf.boneids[3] = readlong app_file #unsigned + 1
			inf.weights[3] = readfloat app_file
			if inf.weights[3] == 0 then (
				deleteitem inf.boneids 3 
				deleteitem inf.weights 3
			)
			if inf.weights[2] == 0 then (
				deleteitem inf.boneids 2 
				deleteitem inf.weights 2
			)
			if inf.weights[1] == 0 then (
				deleteitem inf.boneids 1 
				deleteitem inf.weights 1
			)
			s.infl[j] = inf
		)
		app_file_restore_pos pos app_file
		app_skip_byte 4 app_file	--pad
		num_indices = readlong app_file #unsigned
		s.num_faces = num_indices/3
		s.indices = #()
		s.indices[s.num_faces] = 0
		offset = readlong app_file #unsigned
		size = readlong app_file #unsigned
		pos = app_file_pos offset app_file
		for j = 1 to s.num_faces do(
			face = #()
			face[3] = 0
			face[1] = readshort app_file #unsigned + 1
			face[2] = readshort app_file #unsigned + 1 
			face[3] = readshort app_file #unsigned + 1
			s.indices[j] = face
		)
		app_file_restore_pos pos app_file
		app_skip_byte 4 app_file	--pad
		app_skip_byte 4 app_file	--cloth_struct_offset
		app_skip_byte 4 app_file	--cloth_struct_size
		app_skip_byte 4 app_file	--pad
		local snoSurface = readlong app_file #unsigned
		app_skip_byte 4 app_file	--unknown int
		app_skip_byte 4 app_file	--unknown float
		s.name = app_read_fixed_str 128 app_file
		local cloth_name = app_read_fixed_str 128 app_file
		bb = #()
		bb[6] = 0
		bb[1] = readfloat app_file
		bb[2] = readfloat app_file
		bb[3] = readfloat app_file
		bb[4] = readfloat app_file
		bb[5] = readfloat app_file
		bb[6] = readfloat app_file
		s.bbox = bb
		app_skip_byte 4 app_file	--num_coll_shapes
		app_skip_byte 4 app_file	--col_shapes_offset
		app_skip_byte 4 app_file	--col_shapes_size
		app_skip_byte 8 app_file	--pad
		subs[i] = s
	)
	return subs
)

fn app_read_hardpoints count = (
	hps = #()
	hps[count] = 0
	for i = 1 to count do (
		hp = app_hardpoint()
		hp.name = app_read_fixed_str 64 app_file
		hp.parentID = readlong app_file #unsigned + 1
		rx = readfloat app_file
		ry = readfloat app_file
		rz = readfloat app_file
		rw = readfloat app_file
		px = readfloat app_file
		py = readfloat app_file
		pz = readfloat app_file
		hp.rot = quat rx ry rz rw
		hp.pos = point3 px py pz
--		format "name: %\nparentID: %\npos.x = %; pos.y = %; pos.z = %\nrot.x = %; rot.y = %; rot.z = %; rot.w = %\n" hp.name hp.parentID hp.pos.x hp.pos.y hp.pos.z hp.rot.x hp.rot.y hp.rot.z hp.rot.w 
		hps[i] = hp
	)
	return hps
)

fn app_read =
(
	local deadbeef = readlong app_file #unsigned
	app_skip_byte 28 app_file
    local flags = readlong app_file #unsigned
    num_bones = readlong app_file #unsigned
	tag_offset = readlong app_file #unsigned
	tag_size = readlong app_file #unsigned
	if num_bones != 0 then(
		local pos = app_file_pos tag_offset app_file
		app_read_bones()
		app_file_restore_pos pos app_file
--		app_debug_print_bones()
	)
	app_skip_byte 8 app_file --pad
	app_skip_byte 4 app_file --lookatdata start unknown
	local lookatname = app_read_fixed_str 64 app_file
	local angles = #()
	angles[4] = 0
	angles[1] = readfloat app_file
	angles[2] = readfloat app_file
	angles[3] = readfloat app_file
	angles[4] = readfloat app_file
	local angularvel = readfloat app_file
	format "lookupdata\nname: %\nangles: % % % %\nangular_velocity: %\n" lookatname angles[1] angles[2] angles[3] angles[4] angularvel
	app_skip_byte 4 app_file --lookupdata float unknown
	app_skip_byte 4 app_file --num_bonepulse
	app_skip_byte 8 app_file --pad
	app_skip_byte 4 app_file --bonepulse_offset
	app_skip_byte 4 app_file --bonepulse_size
	--GeoSets
	app_geos = #()
	app_geos[2] = 0
	for i = 1 to 2 do(
		g = app_geoset()
		g.num_subobjects = readlong app_file #unsigned
		num_meshes += g.num_subobjects
		local ofs = readlong app_file #unsigned
		local siz = readlong app_file #unsigned
		if g.num_subobjects != 0 then (
			ppos = app_file_pos ofs app_file
			format "pos before: %\n" pos
			cmeshes = #()
			g.subobjects = app_read_subobjects g.num_subobjects
			app_file_restore_pos ppos app_file
		)
		app_geos[i] = g
		format "pos after: %\n" (ftell app_file)
		format "i=%\n" i 
	)
--	app_file_restore_pos pos app_file
	app_skip_byte 24 app_file -- i don't understand this but file positioning not working properly
	local x = readfloat app_file
	local y = readfloat app_file
	local z = readfloat app_file
	local rad = readfloat app_file
--	sph = sphere pos:(point3 x y z) name:"Bounding Sphere" 
--	sph.radius = rad
--	sph.parent = rbones[1].maxbone
--	sph.wirecolor = (color 6 134 58)
--	format "x %; y %; z %; radius: %\n" x y z rad
	app_skip_byte 4 app_file --num_collision_shapes
	app_skip_byte 4 app_file --offset_collision_shapes
	app_skip_byte 4 app_file --size_collision_shapes
	app_skip_byte 12 app_file --pad
	num_hardpoints = readlong app_file #unsigned
	tag_offset = readlong app_file #unsigned
	tag_size = readlong app_file #unsigned
	if num_hardpoints !=0 then (
		pos = app_file_pos tag_offset app_file
		fhardpoints = app_read_hardpoints num_hardpoints
		app_file_restore_pos pos app_file
	)
)

fn app_create_mesh = (
	for gs = 1 to 2 do(
		g = app_geos[gs]
		if g != 0 then(
			for i = 1 to g.num_subobjects do(
				so = g.subobjects[i]
				v = #()
				t = #()
				for j = 1 to so.num_verts do (
					ru = (so.verts[j].u - 32767)/512.0
					rv = 1 - (so.verts[j].v - 32767)/512.0
--					format "u: %, v: %\n" ru rv
					append v (point3 so.verts[j].x so.verts[j].y so.verts[j].z)
					append t (point3 ru rv 0.0)
				)
				f = #()
				for j = 1 to so.num_faces do (
					append f (point3 so.indices[j][1] so.indices[j][2] so.indices[j][3])
				)
				
				cmesh = mesh vertices:v faces:f name:so.name
				update cmesh
				
				meshop.setnummaps cmesh 1
				meshop.setmapsupport cmesh 1 true
				meshop.defaultmapfaces cmesh 1
				for j = 1 to t.count do (
					meshop.setmapvert cmesh 1 j t[j]
				)
				
				update cmesh
				append cmeshes cmesh
			)
		)
	)
	max views redraw
)

fn app_create_skin = (
	for gs = 1 to 2 do (
		g = app_geos[gs]
		if g != 0 then(
			max modify mode
			for i = 1 to g.num_subobjects do(
				sk = skin name:"Skin"
				cmesh = cmeshes[i]
				
				addModifier cmesh sk
				modpanel.setcurrentobject sk
				
				for j = 1 to num_bones do (
					cb = getnodebyname rbones[j].name
					skinops.addbone sk cb 0
				)
				update cmesh
				
				max views redraw
				disablesceneredraw()
				so = g.subobjects[i]
				for j = 1 to so.num_verts do(
					local inf = so.infl[j]
					skinops.replacevertexweights sk j inf.boneids inf.weights
				)
				enablesceneredraw()
				update cmesh
				redrawviews()
			)
		)
	)
)

fn app_create_hardpoints = (
	if num_hardpoints != 0 then (
		for i = 1 to num_hardpoints do (
			hp = fhardpoints[i]
			cp = point name:hp.name size:0.5
			cp.rotation = hp.rot
			cp.pos = hp.pos
			if hp.parentID != 0 then cp.parent = rbones[hp.parentID].maxbone
		)
	)
)

fn app_boneid_by_name bname = (
	for i = 1 to num_bones do(
		if rbones[i].name == bname then return i
	)
)

fn app_create_bones = (
	cbones = #()
	for i = 1 to num_bones do (
		b = rbones[i]
		cb = BoneSys.CreateBone [0, 0, 0] [0, 0, 0] [0, 0, 0.1]
		cb.name = b.name
		cb.width = 0.025
		cb.height = 0.025
		cb.showlinks = true
		cb.bonescaletype = #none
		cb.rotation = b.trans[1].rot
		cb.position = b.trans[1].pos
		cb.scale = point3 b.trans[1].scale b.trans[1].scale b.trans[1].scale
		b.maxbone = cb
		append cbones cb
	)

	max views redraw
	
	for i = 1 to num_bones do (
		cbones[i].parent = undefined
		if rbones[i].parentID != 0 do (
			cbones[i].parent = cbones[rbones[i].parentID]
		)
	)
	
	for i = 1 to num_bones do(
		b = rbones[i]
		b.maxbone.setskinpose()
		b.bmat = b.maxbone.transform
	)
	max views redraw
)

fn app_main filename =
(
	if filename == undefined then
	(
		print "filename undefined"
		throw "filename undefined"
	)
	app_reset_globals()
	app_file = app_open_file(filename)
	app_read()
	app_close_file(app_file)
	
	if num_bones != 0 then (
		app_create_bones()
	)
	app_create_hardpoints()
	
	app_create_mesh()
	if num_bones != 0 then (
		app_create_skin()
	)
)

-- Animation 
fn app_read_animations count =
(
	
	if g_num_anims == 0 then (
		anims = #()
		anims[count + 1] = 0
	)
	for i = 1 to count do(
		an = animation()
		readlong 4 ani_file
		an.name = app_read_fixed_str 64 ani_file
		app_skip_byte 56 ani_file
		local l_num_bones = an.num_bones = readlong ani_file #unsigned
		local offset = readlong ani_file #unsigned
		local size = readlong ani_file #unsigned
		x = app_file_pos offset ani_file
		an.ani_bones = #()
		an.ani_bones[l_num_bones + 1] = 0
		for j = 1 to l_num_bones do(
			ab = ani_bone()
			ab.name = app_read_fixed_str 64 ani_file
			an.ani_bones[j] = ab
		)
		app_file_restore_pos x ani_file
		app_skip_byte 12 ani_file
		an.num_frames = readlong ani_file #unsigned
		offset = readlong ani_file #unsigned
		size = readlong ani_file #unsigned
		x = app_file_pos offset ani_file
		for j = 1 to l_num_bones do (
			an.ani_bones[j].num_keys_pos = readlong ani_file #unsigned
			an.ani_bones[j].frames_pos = #()
			an.ani_bones[j].frames_pos[an.ani_bones[j].num_keys_pos + 1] = 0
			local ofs = readlong ani_file #unsigned
			local siz = readlong ani_file #unsigned
			xx = app_file_pos ofs ani_file
			for k = 1 to an.ani_bones[j].num_keys_pos do (
				local af = ani_frame_pos()
				af.key = readlong ani_file #unsigned
				af.pos = point3 0.0 0.0 0.0
				af.pos.x = readfloat ani_file
				af.pos.y = readfloat ani_file
				af.pos.z = readfloat ani_file
				an.ani_bones[j].frames_pos[k] = af
			)
			xx += 12
			app_file_restore_pos xx ani_file
		)
		app_file_restore_pos x ani_file
		app_skip_byte 12 ani_file
		offset = readlong ani_file #unsigned
		size = readlong ani_file #unsigned
		x = app_file_pos offset ani_file
		for j = 1 to l_num_bones do (
			an.ani_bones[j].num_keys_rot = readlong ani_file #unsigned
			an.ani_bones[j].frames_rot = #()
			an.ani_bones[j].frames_rot[an.ani_bones[j].num_keys_rot + 1] = 0
			local ofs = readlong ani_file #unsigned
			local siz = readlong ani_file #unsigned
			xx = app_file_pos ofs ani_file
			for k = 1 to an.ani_bones[j].num_keys_rot do (
				local af = ani_frame_rot()
				af.key = readlong ani_file #unsigned
				af.rot = quat 0.0 0.0 0.0 0.0
				local x = 1.0 
				local y = 1.0
				local z = 1.0
				local w = 1.0
				x = readshort ani_file 
				y = readshort ani_file
				z = readshort ani_file
				w = readshort ani_file
				af.rot.x = x / 32767.0
				af.rot.y = y / 32767.0
				af.rot.z = z / 32767.0
				af.rot.w = w / 32767.0
				an.ani_bones[j].frames_rot[k] = af
			)
			xx += 12
			app_file_restore_pos xx ani_file
		)
		app_file_restore_pos x ani_file
		app_skip_byte 8 ani_file
		offset = readlong ani_file #unsigned
		size = readlong ani_file #unsigned
		x = app_file_pos offset ani_file
		for j = 1 to l_num_bones do (
			an.ani_bones[j].num_keys_scale = readlong ani_file #unsigned
			an.ani_bones[j].frames_scale = #()
			an.ani_bones[j].frames_scale[an.ani_bones[j].num_keys_scale + 1] = 0
			local ofs = readlong ani_file #unsigned
			local siz = readlong ani_file #unsigned
			xx = app_file_pos ofs ani_file
			for k = 1 to an.ani_bones[j].num_keys_scale do (
				local af = ani_frame_scale()
				af.key = readlong ani_file #unsigned
				af.scale = readfloat ani_file
				an.ani_bones[j].frames_scale[k] = af
			)
			xx += 12
			app_file_restore_pos xx ani_file
		)
		app_file_restore_pos x ani_file
		app_skip_byte 8 ani_file
		app_skip_byte 64 ani_file		--unknown matrix
		app_skip_byte 4 ani_file		--float -1.0
		app_skip_byte 8 ani_file		--serDataHardPoints
		app_skip_byte 124 ani_file     --unknown field with 2 offset/size references
		local xanNumber = app_read_fixed_str 256 ani_file
		local xanName = app_read_fixed_str 256 ani_file
		app_skip_byte 8 ani_file
		an.applyed = false
		anims[i] = an
	)
)

fn app_apply_anim =
(
	for i = 1 to f_num_anims do (
		an = anims[i]
		if an.applyed != true then (
			for j = 1 to an.num_bones do (
				cba = an.ani_bones[j]
				cb = getnodebyname cba.name
				if cb !=undefined then (
					for k = 1 to cba.num_keys_rot do (
						with animate on(
							at time (cur_frame + cba.frames_rot[k].key) (
								cb.assumeskinpose()
								local m = matrix3 1
								rotate m (inverse cba.frames_rot[k].rot)
								m.row4 = cb.pos
								if cb.parent != undefined then(
									m = m * cb.parent.transform 
								)
								cb.transform = m
								deleteKey cb.position.controller (numKeys cb.position.controller)
								deleteKey cb.scale.controller (numKeys cb.scale.controller)
							)
						)
					)
					for k = 1 to cba.num_keys_pos do (
						with animate on(
							at time (cur_frame + cba.frames_pos[k].key) (
								in coordsys parent cb.position = cba.frames_pos[k].pos
							)
						)
					)
					for k = 1 to cba.num_keys_scale do (
						with animate on(
							at time (cur_frame + cba.frames_scale[k].key) (
								in coordsys parent cb.scale = (point3 cba.frames_scale[k].scale cba.frames_scale[k].scale cba.frames_scale[k].scale)
							)
						)
					)
				)
			)
			
			an.applyed = true
			cur_frame += 2
			cur_frame += an.num_frames
		)
--		for i = 1 to num_bones do (
--			b = rbones[i]
--			b.maxbone.transform = b.bmat
--			addNewKey b.maxbone.rotation.controller (cur_frame - 1)
--			addNewKey b.maxbone.position.controller (cur_frame - 1)
--			addNewKey b.maxbone.scale.controller (cur_frame - 1)
--		)
	)
)

fn app_main_ani filename =
(
	ani_file = app_open_file filename
	local deadbeef = readlong ani_file #undigned
	app_skip_byte 28 ani_file
	local flags = readlong ani_file #unsigned
	local snoAppearance = readlong ani_file #unsigned
	tag_offset = readlong ani_file #unsigned
	tag_size = readlong ani_file #unsigned
	f_num_anims = readlong ani_file #unsigned
	local pos = app_file_pos tag_offset ani_file
	app_read_animations f_num_anims
	g_num_anims += f_num_anims
	app_file_pos pos ani_file
	app_close_file ani_file
	print ("cur_frame = " + (cur_frame as string) + "\n")
	
	app_apply_anim()
	print ("cur_frame = " + (cur_frame as string) + "\n" )
)
--UI
fn uiOpenFile ftype &fname =
(
	local ret = getOpenFileName types:ftype filename:fname
	if ret != undefined then fname = ret
)

utility appimp "App Import"
(
	group "Model filename"
	(
		button bOpenFile "Open..."
		edittext tFileName
		button bImport "Import" height:35 width:100
	)
	
	on bOpenFile pressed do (
		uiOpenFile "App model (*.app)|*.app|All files|*.*" &tFileName.text
	)
	
	on bImport pressed do (
		app_main tFilename.text
	)

	group "Animation filename"
	(
		button bOpenFileAnim "Open..."
		edittext tFileNameAnim
		button bImportAnim "Import" height:35 width:100
	)
	
	on bOpenFileAnim pressed do (
		uiOpenFile "Ani animation (*.ani)|*.ani|All files|*.*" &tFileNameAnim.text
	)
	
	on bImportAnim pressed do (
		app_main_ani tFilenameAnim.text
	)
)

